package ij.plugin.filter;
import ij.*;
import ij.process.*;
import java.awt.*;

/** Implements ImageJ's Equalize command. It changes the tone curves of 
images. It should bring up the detail in the flat regions of your image.
Histogram Equalization can enhance meaningless detail and hide 
important but small high-contrast features. This plugin uses a
similar algorithm, but uses the square root of the histogram 
values, so its effects are less extreme. Hold the alt key down, or set
'arg' to "classic", to use the standard histogram equalization algorithm.
This code was contributed by Richard Kirk (rak@cre.canon.co.uk).
*/
public class Equalizer implements PlugInFilter {

	ImagePlus imp;
	int[] histogram;
	int min, max, offset, range;
	boolean classicEqualization = IJ.altKeyDown();

	/** Set 'arg' to "classic" to use the standard histogram equalization algorithm. */
	public int setup(String arg, ImagePlus imp) {
		if (imp==null)
			return DONE;
		this.imp = imp;
		ImageProcessor ip = imp.getProcessor();
		histogram = ip.getHistogram();
		imp.killRoi();
		if (arg.equals("classic"))
			classicEqualization = true;
		return DOES_8G+DOES_RGB+DOES_16+DOES_STACKS;
	}

	private double getWeightedValue(int i) {
		if (i<offset) return 0.0;
		int h = histogram[i-offset];
		if (h<2 || classicEqualization) return (double)h;
		return Math.sqrt((double)(h));
	}
	
	public void run(ImageProcessor ip) {
	
		if (histogram==null)
			return;
	
		
		if (ip instanceof ShortProcessor) {	// Short
			if (imp.getStackSize()>1)
				return;
			min = (int)ip.getMin();
			max = (int)ip.getMax();
			offset = min;
			range = max - min;
		} else {
			min = 0;	// Bytes
			max = 255;
			offset = 0;
			range = 255;
		}
		
		double sum;
		
		sum = getWeightedValue(min);
		for (int i=min+1; i<max; i++)
			sum += 2 * getWeightedValue(i);
		sum += getWeightedValue(max);
		
		double scale = range/sum;
		int[] lut = new int[range+1];
		
		for (int i=offset; i<=min; i++)
			lut[i-offset] = offset;
		
		sum = getWeightedValue( min );
		for (int i=min+1; i<max; i++) {
			double delta = getWeightedValue(i);
			sum += delta;
			lut[ i-offset ] = offset + (int)Math.round(sum*scale) ;
			sum += delta;
		}
		
		for (int i=max; i<= offset+range; i++)
			lut[ i-offset ] = max;
		
		ip.applyTable(lut);
	}

}
